//************************************************************
//  University of Patras 
//	Department of Electrical Engineering & Computer Technology
//	 
//	Course: Parallel Programming in AI 
//	Week 01: Code Optimization
//
//  Description: Arithmetic operations of Large two dimensional Matrices
//  	addmatOrg02.c: Re-arranging the orders of two loops
//	Author: Evangelos Dermatas
//	v0.1
/*
*********************************************************************
gcc -v:	gcc version 7.4.0 (Ubuntu 7.4.0-1ubuntu1~18.04.1)
OS: 	Linux Ubuntu 18.04
CPU:	8-core, Intel(R) Core(TM) i7-2630QM CPU @ 2.00GHz, 3991 BogoMips
*********************************************************************
Compiling:	gcc addmatOrg02Vect.c -o addmatOrg02Vect -fopt-info-vec-optimized
Executing:	time ./addmatOrg02Vect
Output in my Laptop:
innerProdMat  207115863

real	0m8.946s
user	0m8.942s
sys		0m0.004s
*/
//************************************************************

// *******************************************************************
#pragma GCC optimize("O3","unroll-loops","omit-frame-pointer","inline") 
#pragma GCC option("arch=native","tune=native","no-zero-upper") 
//************************************************************

#include <stdio.h>
#include <stdlib.h>
//#include <omp.h>

// Matrices Dimension
#define N	1000

// **************************************************************
int A[N][N] ;
int B[N][N] ;
int C[N][N] ;

// ***************************************************************************
//    Set random values to the elements of a two dimensional matrix a[N][N] 
// 		in the range [-50,49]
// ***************************************************************************
void setValMat( int a[N][N] ) 
{
	int i, j ;
	for( i = 0 ; i < N ; i++ )
		for( j = 0 ; j < N ; j++ )
		    a[i][j] = (rand() % 100) - 50 ; 
}

// ***************************************************************************
//    Add matrix a to b into c
// ***************************************************************************
void addMat( int a[N][N], int b[N][N], int c[N][N] ) 
{
	int i, j ;
   	for( i = 0 ; i < N ; i++ )
		for( j = 0 ; j < N ; j++ )
	    	c[i][j] = a[i][j] + b[i][j] ; 

}

// ***************************************************************************
//    Compute the inner product of a and b
// ***************************************************************************
int innerProdMat( int a[N][N], int b[N][N] ) 
{
	int i, j, sum = 0 ;
	for( i = 0 ; i < N ; i++ )
	    for( j = 0 ; j < N ; j++ )
		    sum += a[i][j]*b[i][j] ; 
	return sum ;
}

// ***************************************************************************
//    Compute in c the product of a and b
// ***************************************************************************
void mulMat( int a[N][N], int b[N][N], int c[N][N] ) 
{
	int i, j, k ;
   	for( i = 0 ; i < N ; i++ )
		for( j = 0 ; j < N ; j++ )
	        for( k = 0, c[i][j] = 0 ; k < N ; k++ )
		        c[i][j] += a[i][k] * b[k][j] ; 
}

// ***************************************************************************
//    Compute Transpose of a into  b
// ***************************************************************************
void tranMat( int a[N][N], int b[N][N] ) 
{
	int i, j ;
	
   	for( i = 0 ; i < N ; i++ )
		for( j = 0 ; j < N ; j++ )
			b[i][j] = a[j][i] ;
}

// ***************************************************************************
//    Compute in c the product of a and b
// ***************************************************************************
void mulMat01( int a[N][N], int b[N][N], int c[N][N] ) 
{
	int i, j, k, s = 0 ;
	int T[N][N] ;

	tranMat( b, T) ; 
   	for( i = 0 ; i < N ; i++ ) 
		for( j = 0 ; j < N ; j++ ) {
	        for( k = s = 0 ; k < N ; k++ )
		        s += a[i][k] * T[j][k] ;
		c[i][j] = s ;
	   } 
}


// **************************************************************
int main(void) 
{
//double t1, t2, ttot = 0, t ;
int res ;
//	t1 = omp_get_wtime() ;
	setValMat( A ) ;
//	t2 = omp_get_wtime() ;
//	ttot = t = t2-t1 ;
//	printf( "SetVal A      %lf\n", t ) ;

//	t1 = omp_get_wtime() ;
	setValMat( B ) ;
//	t2 = omp_get_wtime() ;
//	ttot += t = t2-t1 ;
//	printf( "SetVal B      %lf\n", t ) ;

//	t1 = omp_get_wtime() ;
	addMat( A, B, C ) ; 
//	t2 = omp_get_wtime() ;
//	ttot += t = t2-t1 ;
//	printf( "addMat        %lf\n", t ) ;

//	t1 = omp_get_wtime() ;
	mulMat01( C, B, A ) ; 
//	t2 = omp_get_wtime() ;
//	ttot += t = t2-t1 ;
//	printf( "MulMat        %lf\n", t) ;

//	t1 = omp_get_wtime() ;
	res = innerProdMat( A, B ) ; 
//	t2 = omp_get_wtime() ;
//	ttot += t = t2-t1 ;
//	printf( "innerProdMat  %lf %d\n", t, res ) ;
	printf( "innerProdMat  %d\n", res ) ;

//	printf( "Total Time:   %lf\n", ttot) ;
	return 0 ;
}

