//************************************************************
//  University of Patras 
//	Department of Electrical Engineering & Computer Technology
//	 
//	Course: Parallel Programming in AI 
//	Week 02: Simple parallel programming methods
//
//  Problem: How many prime numbers are in [I1,I2) ? 
//  	primes03.c: Sieve of Eratosthenes (Η σίτα του Ερατοσθένη)
//	Author: Evangelos Dermatas
//	v0.1
/*
*********************************************************************
g++ -v:	gcc version 7.4.0 (Ubuntu 7.4.0-1ubuntu1~18.04.1)
OS: 	Linux Ubuntu 18.04
CPU:	8-core, Intel(R) Core(TM) i7-2630QM CPU @ 2.00GHz, 3991 BogoMips
*********************************************************************
Compiling:	gcc primes03.c -o primes03 -std=c99
Executing:	time ./primes03 3 10000000 
Output in my laptop:
Number of Primes in [3,10000000) is 664578

real	0m0.179s
user	0m0.175s
sys	    0m0.004s
*********************************************************************
Compiling:	gcc primes03.c -o primes03 -std=c99 -O2
Executing:	time ./primes03 3 10000000 
Output in my laptop:
Number of Primes in [3,10000000) is 664578

real	0m0.084s
user	0m0.084s
sys	    0m0.000s
*********************************************************************
Compiling:	icc primes03.c -o primes03 -std=c99 -O2
Executing:	time ./primes03 3 10000000 
Output in my laptop:
Number of Primes in [3,10000000) is 664578

real	0m0.087s
user	0m0.079s
sys 	0m0.008s
*********************************************************************
*/

// *******************************************************************
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

// **********************************************************
int primes( int min, int Max ) 
{ 
    bool *P = calloc(sizeof(bool),Max); 
    if ( P == NULL ) return -1 ;
  
    // Build the Sieve of Eratosthenes 
    for (int p = 2; p*p < Max; p++) { 
          if (P[p] == false) { 
            for (int i = p*2; i < Max; i += p) 
                P[i] = true; 
        } 
    } 
  
    // Count in the Sieve of Eratosthenes 
    int s = 0 ; 
    for (int i = (min|1) ; i < Max; i += 2) { 
        if ( P[i] == false ) 
            s++; 
    } 
    free(P) ;
    return s ;
} 

// *******************************************************************
int main(int argc, char *argv[]) 
{ 
    int m, M, res ;

    if ( argc < 3)
        return -1 ;
    else {
        if ((m=atoi(argv[1]))<3 )
            return -2 ;
        if ((M=atoi(argv[2]))<3 )
            return -2 ;
    }
    res = primes(m,M) ;
    printf( "Number of Primes in [%d,%d) is %d\n", m, M, res ) ;

	return res; 
} 
